#!/usr/bin/python3

from argparse import ArgumentParser
from contextlib import suppress
import os
import re
import shlex
import shutil
import subprocess as sp
import sys

def kexec_main(a):
    unshare_env = f'reboot_unshared_{os.getpid()}'
    if unshare_env not in os.environ:
        os.environ[unshare_env] = '1'
        c = ['unshare', '-m', sys.executable] + sys.argv
        os.execvp(c[0], c)

    def listentry():
        for e in os.listdir(os.path.join(a.boot, 'loader', 'entries')):
            if e.endswith('.conf'):
                yield e.rsplit('.', 1)[0]

    def loadentry(e):
        r = {}
        with open(os.path.join(a.boot, 'loader', 'entries', e + '.conf')) as f:
            for l in f:
                l = l.split(' ', 1)
                if len(l) == 2:
                    r[l[0]] = l[1].strip()
        return r

    entry = a.boot_entry
    if entry is None:
        entry = sorted(listentry(), key=lambda x: tuple(re.findall(r'\d+', x)))[-1]

    entry = loadentry(entry)
    version = entry['version']
    kernel = entry['linux'].lstrip('/')
    option = entry.get('options', '') + ' rd.shell=0 rd.break=0 rd.timeout=300 rd.emergency=reboot'
    print(f'Selected kernel {kernel} version {version}')

    dracutcmd = ['dracut', '--no-compress']
    if 'rd.luks.uuid' in option and os.path.exists('/etc/crypttab'):
        luksuuid = []
        for o in option.split():
            if o.startswith('rd.luks.uuid='):
                u = o.split('=', 1)[1]
                if u.startswith('keysource:'):
                    u = u.split(':', 1)[1]
                if u.startswith('luks-'):
                    u = u[5:]
                luksuuid.append(u)
        keyfiles = set()
        with open('/etc/crypttab') as f:
            for l in f:
                l = l.split()
                if len(l) >= 3 and l[1].startswith('UUID=') and l[2].startswith('/'):
                    uuid = l[1][5:]
                    for u in luksuuid:
                        if uuid.startswith(u):
                            keyfiles.add(l[2])
        if keyfiles:
            dracutcmd += ['--install', '/etc/crypttab']
            for f in keyfiles:
                dracutcmd += ['--install', f]

    dracutcmd += ['/tmp/initrd.img', version]
    sp.run(['mount', '-t', 'tmpfs', 'none', '/tmp'], stdin=sp.DEVNULL, check=True)
    sp.run(dracutcmd, stdin=sp.DEVNULL, check=True)

    sp.run(['kexec', '-l', '-a', '--append=' + option, '--initrd=/tmp/initrd.img', os.path.join(a.boot, kernel)], stdin=sp.DEVNULL, check=True)

    if not a.load:
        sp.run(['systemctl', '--no-block', 'kexec'], stdin=sp.DEVNULL, check=True)

def main():
    a = ArgumentParser(description='reboot by kexec')
    a.add_argument('--load', '-l', action='store_true', help='load kernel but do not reboot')
    a.add_argument('--boot', default='/boot', help='specify $BOOT partition mount point')
    a.add_argument('boot_entry', nargs='?', help='select boot entry')

    kexec_main(a.parse_args())

if __name__ == '__main__':
    main()
